/*
Copyright 2025 The Kubernetes Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

package fieldpath_test

import (
	"sigs.k8s.io/structured-merge-diff/v6/fieldpath"
	"sigs.k8s.io/structured-merge-diff/v6/value"
)

// Type and function aliases to avoid typing...

type (
	Filter      = fieldpath.Filter
	Path        = fieldpath.Path
	PathElement = fieldpath.PathElement
	Set         = fieldpath.Set
	SetNodeMap  = fieldpath.SetNodeMap
)

var (
	KeyByFields             = fieldpath.KeyByFields
	MakePathOrDie           = fieldpath.MakePathOrDie
	MakePrefixMatcherOrDie  = fieldpath.MakePrefixMatcherOrDie
	MatchAnyPathElement     = fieldpath.MatchAnyPathElement
	NewIncludeMatcherFilter = fieldpath.NewIncludeMatcherFilter
	NewSet                  = fieldpath.NewSet

	// Short names for readable test cases.
	_V  = value.NewValueInterface
	_NS = fieldpath.NewSet
	_P  = fieldpath.MakePathOrDie
)
